@@ -1,22 +1,36 @@
 """Identifiable SDK client."""
 
+import concurrent.futures
 import io
 import os
 from pathlib import Path
-from typing import Any, cast
+from typing import Any, TypeVar, cast
 
 import httpx
 
 from entitysdk import core, route
 from entitysdk.common import ProjectContext
 from entitysdk.exception import EntitySDKError
-from entitysdk.models.asset import Asset, DetailedFileList, LocalAssetMetadata
+from entitysdk.models.asset import (
+    Asset,
+    DetailedFileList,
+    ExistingAssetMetadata,
+    LocalAssetMetadata,
+)
 from entitysdk.models.core import Identifiable
 from entitysdk.models.entity import Entity
 from entitysdk.result import IteratorResult
-from entitysdk.schemas.asset import DownloadedAsset
+from entitysdk.schemas.asset import DownloadedAssetFile
 from entitysdk.token_manager import TokenFromValue, TokenManager
-from entitysdk.types import ID, DeploymentEnvironment, Token
+from entitysdk.types import (
+    ID,
+    AssetLabel,
+    ContentType,
+    DeploymentEnvironment,
+    DerivationType,
+    StorageType,
+    Token,
+)
 from entitysdk.util import (
     build_api_url,
     create_intermediate_directories,
@@ -24,6 +38,8 @@ from entitysdk.util import (
 )
 from entitysdk.utils.asset import filter_assets
 
+TIdentifiable = TypeVar("TIdentifiable", bound=Identifiable)
+
 
 class Client:
     """Client for entitysdk."""
@@ -94,9 +110,9 @@ class Client:
         self,
         entity_id: ID,
         *,
-        entity_type: type[Identifiable],
+        entity_type: type[TIdentifiable],
         project_context: ProjectContext | None = None,
-    ) -> Identifiable:
+    ) -> TIdentifiable:
         """Get entity from resource id.
 
         Args:
@@ -124,11 +140,11 @@ class Client:
     def search_entity(
         self,
         *,
-        entity_type: type[Identifiable],
+        entity_type: type[TIdentifiable],
         query: dict | None = None,
         limit: int | None = None,
         project_context: ProjectContext | None = None,
-    ) -> IteratorResult[Identifiable]:
+    ) -> IteratorResult[TIdentifiable]:
         """Search for entities.
 
         Args:
@@ -149,12 +165,30 @@ class Client:
             token=self._token_manager.get_token(),
         )
 
+    def get_entity_derivations(
+        self,
+        *,
+        entity_id: ID,
+        entity_type: type[Entity],
+        derivation_type: DerivationType,
+        project_context: ProjectContext | None = None,
+    ) -> IteratorResult[Entity]:
+        """Get all the derivation for an entity."""
+        return core.get_entity_derivations(
+            api_url=self.api_url,
+            entity_id=entity_id,
+            entity_type=entity_type,
+            derivation_type=derivation_type,
+            project_context=self._required_user_context(override_context=project_context),
+            token=self._token_manager.get_token(),
+        )
+
     def register_entity(
         self,
-        entity: Identifiable,
+        entity: TIdentifiable,
         *,
         project_context: ProjectContext | None = None,
-    ) -> Identifiable:
+    ) -> TIdentifiable:
         """Register entity.
 
         Args:
@@ -177,11 +211,11 @@ class Client:
     def update_entity(
         self,
         entity_id: ID,
-        entity_type: type[Identifiable],
+        entity_type: type[TIdentifiable],
         attrs_or_entity: dict | Identifiable,
         *,
         project_context: ProjectContext | None = None,
-    ) -> Identifiable:
+    ) -> TIdentifiable:
         """Update an entity.
 
         Args:
@@ -211,9 +245,10 @@ class Client:
         entity_id: ID,
         entity_type: type[Identifiable],
         file_path: os.PathLike,
-        file_content_type: str,
+        file_content_type: ContentType,
         file_name: str | None = None,
         file_metadata: dict | None = None,
+        asset_label: AssetLabel,
         project_context: ProjectContext | None = None,
     ) -> Asset:
         """Upload asset to an existing entity's endpoint from a file path."""
@@ -229,6 +264,7 @@ class Client:
             file_name=file_name or path.name,
             content_type=file_content_type,
             metadata=file_metadata,
+            label=asset_label,
         )
         return core.upload_asset_file(
             url=url,
@@ -246,8 +282,9 @@ class Client:
         entity_type: type[Identifiable],
         file_content: io.BufferedIOBase,
         file_name: str,
-        file_content_type: str,
+        file_content_type: ContentType,
         file_metadata: dict | None = None,
+        asset_label: AssetLabel,
         project_context: ProjectContext | None = None,
     ) -> Asset:
         """Upload asset to an existing entity's endpoint from a file-like object."""
@@ -261,6 +298,7 @@ class Client:
             file_name=file_name,
             content_type=file_content_type,
             metadata=file_metadata or {},
+            label=asset_label,
         )
         context = self._required_user_context(override_context=project_context)
         return core.upload_asset_content(
@@ -280,7 +318,7 @@ class Client:
         name: str,
         paths: dict[os.PathLike, os.PathLike],
         metadata: dict | None = None,
-        label: str | None = None,
+        label: AssetLabel,
         project_context: ProjectContext | None = None,
     ) -> Asset:
         """Attach directory to an entity from with a group of paths."""
@@ -343,8 +381,9 @@ class Client:
         output_path: os.PathLike,
         project_context: ProjectContext | None = None,
         ignore_directory_name: bool = False,
+        max_concurrent: int = 1,
     ) -> list[Path]:
-        """List directory existing entity's endpoint from a directory path."""
+        """Download directory of assets."""
         output_path = Path(output_path)
 
         if output_path.exists() and output_path.is_file():
@@ -353,34 +392,35 @@ class Client:
 
         context = self._optional_user_context(override_context=project_context)
 
-        asset = None
+        asset = cast(Asset, asset_id) if isinstance(asset_id, Asset) else None
+
         if not ignore_directory_name:
-            asset_endpoint = route.get_assets_endpoint(
-                api_url=self.api_url,
-                entity_type=entity_type,
-                entity_id=entity_id,
-                asset_id=asset_id,
-            )
-            asset = core.get_entity(
-                asset_endpoint,
-                entity_type=Asset,
-                project_context=context,
-                http_client=self._http_client,
-                token=self._token_manager.get_token(),
-            )
+            if asset is None:
+                asset_endpoint = route.get_assets_endpoint(
+                    api_url=self.api_url,
+                    entity_type=entity_type,
+                    entity_id=cast(ID, entity_id),
+                    asset_id=asset_id,
+                )
+                asset = core.get_entity(
+                    asset_endpoint,
+                    entity_type=Asset,
+                    project_context=context,
+                    http_client=self._http_client,
+                    token=self._token_manager.get_token(),
+                )
 
             output_path /= asset.path
 
         contents = self.list_directory(
             entity_id=entity_id,
             entity_type=entity_type,
-            asset_id=asset_id,
+            asset_id=asset_id if isinstance(asset_id, ID) else asset.id,
             project_context=project_context,
         )
 
-        paths = []
-        for path in contents.files:
-            paths.append(
+        if max_concurrent == 1:
+            paths = [
                 self.download_file(
                     entity_id=entity_id,
                     entity_type=entity_type,
@@ -389,7 +429,24 @@ class Client:
                     asset_path=path,
                     project_context=context,
                 )
-            )
+                for path in contents.files
+            ]
+        else:
+            with concurrent.futures.ThreadPoolExecutor(max_workers=max_concurrent) as executor:
+                futures = [
+                    executor.submit(
+                        self.download_file,
+                        entity_id=entity_id,
+                        entity_type=entity_type,
+                        asset_id=asset if asset else asset_id,
+                        output_path=output_path / path,
+                        asset_path=path,
+                        project_context=context,
+                    )
+                    for path in contents.files
+                ]
+                result = concurrent.futures.wait(futures)
+                paths = [res.result() for res in result.done]
 
         return paths
 
@@ -399,6 +456,7 @@ class Client:
         entity_id: ID,
         entity_type: type[Identifiable],
         asset_id: ID,
+        asset_path: os.PathLike | None = None,
         project_context: ProjectContext | None = None,
     ) -> bytes:
         """Download asset content.
@@ -407,6 +465,7 @@ class Client:
             entity_id: Id of the entity.
             entity_type: Type of the entity.
             asset_id: Id of the asset.
+            asset_path: for asset directories, the path within the directory to the file.
             project_context: Optional project context.
 
         Returns:
@@ -425,6 +484,7 @@ class Client:
         return core.download_asset_content(
             url=url,
             project_context=context,
+            asset_path=asset_path,
             http_client=self._http_client,
             token=self._token_manager.get_token(),
         )
@@ -446,13 +506,14 @@ class Client:
             entity_type: Type of the entity.
             asset_id: Id of the asset.
             output_path: Either be a file path to write the file to or an output directory.
-            asset_path: for asset directories, the path within the directory to the file
+            asset_path: for asset directories, the path within the directory to the file.
             project_context: Optional project context.
 
         Returns:
             Output file path.
         """
         context = self._optional_user_context(override_context=project_context)
+
         asset_endpoint = route.get_assets_endpoint(
             api_url=self.api_url,
             entity_type=entity_type,
@@ -494,6 +555,11 @@ class Client:
             token=self._token_manager.get_token(),
         )
 
+    @staticmethod
+    def select_assets(entity: Entity, selection: dict) -> IteratorResult:
+        """Select assets from entity based on selection."""
+        return IteratorResult(filter_assets(entity.assets, selection))
+
     def download_assets(
         self,
         entity_or_id: Entity | tuple[ID, type[Entity]],
@@ -516,9 +582,9 @@ class Client:
                     project_context=context,
                 )
 
-            return DownloadedAsset(
+            return DownloadedAssetFile(
                 asset=asset,
-                output_path=path,
+                path=path,
             )
 
         context = self._optional_user_context(override_context=project_context)
@@ -574,7 +640,7 @@ class Client:
         entity_type: type[Identifiable],
         asset_id: ID,
         file_path: os.PathLike,
-        file_content_type: str,
+        file_content_type: ContentType,
         file_name: str | None = None,
         file_metadata: dict | None = None,
         project_context: ProjectContext | None = None,
@@ -583,7 +649,7 @@ class Client:
 
         Note: This operation is not atomic. Deletion can succeed and upload can fail.
         """
-        self.delete_asset(
+        deleted_asset = self.delete_asset(
             entity_id=entity_id,
             entity_type=entity_type,
             asset_id=asset_id,
@@ -597,4 +663,45 @@ class Client:
             file_name=file_name,
             file_metadata=file_metadata,
             project_context=project_context,
+            asset_label=deleted_asset.label,
+        )
+
+    def register_asset(
+        self,
+        *,
+        entity_id: ID,
+        entity_type: type[Identifiable],
+        name: str,
+        storage_path: str,
+        storage_type: StorageType,
+        is_directory: bool,
+        content_type: ContentType,
+        asset_label: AssetLabel,
+        project_context: ProjectContext | None = None,
+    ) -> Asset:
+        """Register a file or directory already existing."""
+        url = (
+            route.get_assets_endpoint(
+                api_url=self.api_url,
+                entity_type=entity_type,
+                entity_id=entity_id,
+                asset_id=None,
+            )
+            + "/register"
+        )
+        asset_metadata = ExistingAssetMetadata(
+            path=name,
+            full_path=storage_path,
+            storage_type=storage_type,
+            is_directory=is_directory,
+            content_type=content_type,
+            label=asset_label,
+        )
+        context = self._required_user_context(override_context=project_context)
+        return core.register_asset(
+            url=url,
+            project_context=context,
+            asset_metadata=asset_metadata,
+            http_client=self._http_client,
+            token=self._token_manager.get_token(),
         )
