@@ -1,6 +1,7 @@
 """Core SDK operations."""
 
 import io
+import logging
 import os
 from collections.abc import Iterator
 from pathlib import Path
@@ -10,24 +11,35 @@ import httpx
 
 from entitysdk import serdes
 from entitysdk.common import ProjectContext
-from entitysdk.models.asset import Asset, DetailedFileList, LocalAssetMetadata
+from entitysdk.exception import EntitySDKError
+from entitysdk.models.asset import (
+    Asset,
+    DetailedFileList,
+    ExistingAssetMetadata,
+    LocalAssetMetadata,
+)
 from entitysdk.models.core import Identifiable
+from entitysdk.models.entity import Entity
 from entitysdk.result import IteratorResult
+from entitysdk.route import get_entity_derivations_endpoint
+from entitysdk.types import ID, AssetLabel, DerivationType
 from entitysdk.util import make_db_api_request, stream_paginated_request
 
+L = logging.getLogger(__name__)
+
 TIdentifiable = TypeVar("TIdentifiable", bound=Identifiable)
 
 
 def search_entities(
     url: str,
     *,
-    entity_type: type[Identifiable],
+    entity_type: type[TIdentifiable],
     query: dict | None = None,
     limit: int | None,
     project_context: ProjectContext | None = None,
     token: str,
     http_client: httpx.Client | None = None,
-) -> IteratorResult[Identifiable]:
+) -> IteratorResult[TIdentifiable]:
     """Search for entities.
 
     Args:
@@ -77,14 +89,46 @@ def get_entity(
     return serdes.deserialize_model(response.json(), entity_type)
 
 
+def get_entity_derivations(
+    *,
+    api_url: str,
+    entity_id: ID,
+    entity_type: type[Entity],
+    project_context: ProjectContext,
+    derivation_type: DerivationType,
+    token: str,
+    http_client: httpx.Client | None = None,
+) -> IteratorResult[Entity]:
+    """Get derivations for entity."""
+    url = get_entity_derivations_endpoint(
+        api_url=api_url,
+        entity_type=entity_type,
+        entity_id=entity_id,
+    )
+
+    params = {"derivation_type": DerivationType(derivation_type)}
+
+    response = make_db_api_request(
+        url=url,
+        method="GET",
+        project_context=project_context,
+        token=token,
+        http_client=http_client,
+        parameters=params,
+    )
+    return IteratorResult(
+        serdes.deserialize_model(json_data, Entity) for json_data in response.json()["data"]
+    )
+
+
 def register_entity(
     url: str,
     *,
-    entity: Identifiable,
+    entity: TIdentifiable,
     project_context: ProjectContext,
     token: str,
     http_client: httpx.Client | None = None,
-) -> Identifiable:
+) -> TIdentifiable:
     """Register entity."""
     json_data = serdes.serialize_model(entity)
 
@@ -102,12 +146,12 @@ def register_entity(
 def update_entity(
     url: str,
     *,
-    entity_type: type[Identifiable],
+    entity_type: type[TIdentifiable],
     attrs_or_entity: dict | Identifiable,
     project_context: ProjectContext,
     token: str,
     http_client: httpx.Client | None = None,
-) -> Identifiable:
+) -> TIdentifiable:
     """Update entity."""
     if isinstance(attrs_or_entity, dict):
         json_data = serdes.serialize_dict(attrs_or_entity)
@@ -170,6 +214,7 @@ def upload_asset_content(
         url=url,
         method="POST",
         files=files,
+        data={"label": asset_metadata.label} if asset_metadata.label else None,
         project_context=project_context,
         token=token,
         http_client=http_client,
@@ -183,7 +228,7 @@ def upload_asset_directory(
     name: str,
     paths: dict[Path, Path],
     metadata: dict | None = None,
-    label: str | None = None,
+    label: AssetLabel,
     project_context: ProjectContext,
     token: str,
     http_client: httpx.Client | None = None,
@@ -192,7 +237,7 @@ def upload_asset_directory(
     for concrete_path in paths.values():
         if not concrete_path.exists():
             msg = f"Path {concrete_path} does not exist"
-            raise Exception(msg)
+            raise EntitySDKError(msg)
 
     response = make_db_api_request(
         url=url,
@@ -214,11 +259,20 @@ def upload_asset_directory(
         failed = {}
         for path, url in to_upload.items():
             with open(paths[Path(path)], "rb") as fd:
-                response = http_client.request(
-                    method="PUT", url=url, content=fd, follow_redirects=True
-                )
-            if response.status_code != 200:
-                failed[path] = url
+                try:
+                    response = http_client.request(
+                        method="PUT",
+                        url=url,
+                        content=fd,
+                        follow_redirects=True,
+                        timeout=20,
+                    )
+                except httpx.HTTPError:
+                    L.exception("Upload failed, will retry again")
+                    failed[path] = url
+                else:
+                    if response.status_code != 200:
+                        failed[path] = url
         return failed
 
     to_upload = js["files"]
@@ -228,7 +282,7 @@ def upload_asset_directory(
             break
 
     if to_upload:
-        raise Exception(f"Uploading these files failed: {to_upload}")
+        raise EntitySDKError(f"Uploading these files failed: {to_upload}")
 
     return serdes.deserialize_model(js["asset"], Asset)
 
@@ -331,3 +385,23 @@ def delete_asset(
         http_client=http_client,
     )
     return serdes.deserialize_model(response.json(), Asset)
+
+
+def register_asset(
+    url: str,
+    *,
+    asset_metadata: ExistingAssetMetadata,
+    project_context: ProjectContext,
+    token: str,
+    http_client: httpx.Client | None = None,
+) -> Asset:
+    """Register a file or directory already existing."""
+    response = make_db_api_request(
+        url=url,
+        method="POST",
+        json=asset_metadata.model_dump(),
+        project_context=project_context,
+        token=token,
+        http_client=http_client,
+    )
+    return serdes.deserialize_model(response.json(), Asset)
