@@ -8,4 +8,4 @@ from entitysdk.models.asset import Asset
 class HasAssets(BaseModel):
     """Mixin class for entities that have assets."""
 
-    assets: list[Asset] | None = None
+    assets: list[Asset] = []
