@@ -2,6 +2,7 @@
 
 from entitysdk.exception import RouteNotFoundError
 from entitysdk.models.core import Identifiable
+from entitysdk.models.entity import Entity
 from entitysdk.types import ID
 
 # Mapping of entity type to api route name.
@@ -9,14 +10,24 @@ _ROUTES = {
     "Activity": "activity",
     "BrainLocation": "brain-location",
     "BrainRegion": "brain-region",
+    "BrainRegionHierarchy": "brain-region-hierarchy",
+    "CellMorphology": "cell-morphology",
     "Circuit": "circuit",
+    "Consortium": "consortium",
     "Contribution": "contribution",
+    "Derivation": "derivation",
     "ElectricalCellRecording": "electrical-cell-recording",
+    "ElectricalRecordingStimulus": "electrical-recording-stimulus",
+    "EMCellMesh": "em-cell-mesh",
+    "EMDenseReconstructionDataset": "em-dense-reconstruction-dataset",
     "EModel": "emodel",
     "Entity": "entity",
     "ETypeClassification": "etype-classification",
+    "ETypeClass": "etype",
     "Ion": "ion",
+    "IonChannel": "ion-channel",
     "IonChannelModel": "ion-channel-model",
+    "IonChannelRecording": "ion-channel-recording",
     "License": "license",
     "MEModel": "memodel",
     "MEModelCalibrationResult": "memodel-calibration-result",
@@ -24,7 +35,8 @@ _ROUTES = {
     "MTypeClass": "mtype",
     "Organization": "organization",
     "Person": "person",
-    "ReconstructionMorphology": "reconstruction-morphology",
+    "Publication": "publication",
+    "ScientificArtifactPublicationLink": "scientific-artifact-publication-link",
     "Simulation": "simulation",
     "SimulationCampaign": "simulation-campaign",
     "SingleNeuronSimulation": "single-neuron-simulation",
@@ -39,6 +51,13 @@ _ROUTES = {
     "Subject": "subject",
     "Taxonomy": "taxonomy",
     "ValidationResult": "validation-result",
+    # CellMorphologyProtocol type for retrieving
+    "CellMorphologyProtocol": "cell-morphology-protocol",
+    # CellMorphologyProtocol types for registering
+    "DigitalReconstructionCellMorphologyProtocol": "cell-morphology-protocol",
+    "ModifiedReconstructionCellMorphologyProtocol": "cell-morphology-protocol",
+    "ComputationallySynthesizedCellMorphologyProtocol": "cell-morphology-protocol",
+    "PlaceholderCellMorphologyProtocol": "cell-morphology-protocol",
 }
 
 
@@ -67,6 +86,17 @@ def get_entities_endpoint(
     return f"{api_url}/{endpoint}"
 
 
+def get_entity_derivations_endpoint(
+    *,
+    api_url: str,
+    entity_type: type[Entity],
+    entity_id: ID,
+):
+    """Get endpoint for entity derivations."""
+    route_name = get_route_name(entity_type)
+    return f"{api_url}/{route_name}/{entity_id}/derived-from"
+
+
 def get_assets_endpoint(
     *,
     api_url: str,
