@@ -4,16 +4,23 @@ import uuid
 from pathlib import Path
 from unittest.mock import patch
 
+import httpx
 import pytest
 
 from entitysdk.client import Client
 from entitysdk.config import settings
 from entitysdk.exception import EntitySDKError
-from entitysdk.models import Asset, MTypeClass
+from entitysdk.models import Asset, Circuit, MTypeClass
 from entitysdk.models.asset import DetailedFile, DetailedFileList
 from entitysdk.models.core import Identifiable
 from entitysdk.models.entity import Entity
-from entitysdk.types import DeploymentEnvironment
+from entitysdk.types import (
+    AssetLabel,
+    ContentType,
+    DeploymentEnvironment,
+    DerivationType,
+    StorageType,
+)
 
 
 def test_client_api_url():
@@ -58,8 +65,8 @@ def test_client_search(client, httpx_mock):
         method="GET",
         json={
             "data": [
-                {"id": str(id1), "name": "foo", "description": "bar", "type": "zee"},
-                {"id": str(id2), "name": "foo", "description": "bar", "type": "zee"},
+                {"id": str(id1), "name": "foo", "description": "bar", "type": "circuit"},
+                {"id": str(id2), "name": "foo", "description": "bar", "type": "circuit"},
             ],
             "pagination": {"page": 1, "page_size": 10, "total_items": 2},
         },
@@ -128,6 +135,7 @@ def _mock_asset_response(asset_id):
         "status": "created",
         "meta": {},
         "sha256_digest": "sha256_digest",
+        "label": "morphology",
     }
 
 
@@ -145,6 +153,8 @@ def test_client_upload_file(
         method="POST",
         url=f"{api_url}/entity/{entity_id}/assets",
         match_headers=request_headers,
+        match_files={"file": ("foo", b"foo", "application/swc")},
+        match_data={"label": "morphology"},
         json=_mock_asset_response(asset_id),
     )
 
@@ -156,8 +166,9 @@ def test_client_upload_file(
         entity_type=Entity,
         file_name="foo",
         file_path=path,
-        file_content_type="text/plain",
+        file_content_type="application/swc",
         file_metadata={"key": "value"},
+        asset_label="morphology",
     )
 
     assert res.id == asset_id
@@ -172,22 +183,18 @@ def test_client_upload_content(client, httpx_mock, api_url, request_headers):
         method="POST",
         url=f"{api_url}/entity/{entity_id}/assets",
         match_headers=request_headers,
-        match_files={
-            "file": (
-                "foo.txt",
-                buffer,
-                "text/plain",
-            )
-        },
+        match_files={"file": ("foo.swc", buffer, "application/swc")},
+        match_data={"label": "morphology"},
         json=_mock_asset_response(asset_id),
     )
     res = client.upload_content(
         entity_id=entity_id,
         entity_type=Entity,
-        file_name="foo.txt",
+        file_name="foo.swc",
         file_content=buffer,
-        file_content_type="text/plain",
+        file_content_type="application/swc",
         file_metadata={"key": "value"},
+        asset_label="morphology",
     )
 
     assert res.id == asset_id
@@ -212,6 +219,32 @@ def test_client_download_content(client, httpx_mock, api_url, request_headers):
     assert res == b"foo"
 
 
+def test_client_download_content__asset_path(
+    client,
+    httpx_mock,
+    api_url,
+    request_headers,
+):
+    entity_id = uuid.uuid4()
+    asset_id = uuid.uuid4()
+
+    # for downloading the asset
+    httpx_mock.add_response(
+        method="GET",
+        url=f"{api_url}/entity/{entity_id}/assets/{asset_id}/download?asset_path=foo.txt",
+        match_headers=request_headers,
+        content=b"foo",
+    )
+
+    res = client.download_content(
+        entity_id=entity_id,
+        entity_type=Entity,
+        asset_id=asset_id,
+        asset_path="foo.txt",
+    )
+    assert res == b"foo"
+
+
 def test_client_download_file__output_file(
     tmp_path,
     client,
@@ -435,7 +468,7 @@ def test_client_get(
             "id": str(entity_id),
             "name": "foo",
             "description": "bar",
-            "type": "entity",
+            "type": "circuit",
             "assets": [
                 _mock_asset_response(asset_id1),
                 _mock_asset_response(asset_id2),
@@ -456,9 +489,10 @@ def test_client_get(
 def _mock_asset_delete_response(asset_id):
     return {
         "path": "buffer.h5",
-        "full_path": "private/103d7868/103d7868/assets/reconstruction_morphology/8703/buffer.h5",
+        "full_path": "private/103d7868/103d7868/assets/cell_morphology/8703/buffer.swc",
         "is_directory": False,
         "content_type": "application/swc",
+        "label": "morphology",
         "size": 18,
         "sha256_digest": "47ddc1b6e05dcbfbd2db9dcec4a49d83c6f9f10ad595649bacdcb629671fd954",
         "meta": {},
@@ -475,14 +509,14 @@ def test_client_delete_asset(
     api_url,
     request_headers,
 ):
-    mock_route.return_value = "reconstruction-morphology"
+    mock_route.return_value = "cell-morphology"
 
     entity_id = uuid.uuid4()
     asset_id = uuid.uuid4()
 
     httpx_mock.add_response(
         method="DELETE",
-        url=f"{api_url}/reconstruction-morphology/{entity_id}/assets/{asset_id}",
+        url=f"{api_url}/cell-morphology/{entity_id}/assets/{asset_id}",
         match_headers=request_headers,
         json=_mock_asset_delete_response(asset_id),
     )
@@ -506,20 +540,20 @@ def test_client_update_asset(
     api_url,
     request_headers,
 ):
-    mock_route.return_value = "reconstruction-morphology"
+    mock_route.return_value = "cell-morphology"
 
     entity_id = uuid.uuid4()
     asset_id = uuid.uuid4()
 
     httpx_mock.add_response(
         method="DELETE",
-        url=f"{api_url}/reconstruction-morphology/{entity_id}/assets/{asset_id}",
+        url=f"{api_url}/cell-morphology/{entity_id}/assets/{asset_id}",
         match_headers=request_headers,
         json=_mock_asset_response(asset_id),
     )
     httpx_mock.add_response(
         method="POST",
-        url=f"{api_url}/reconstruction-morphology/{entity_id}/assets",
+        url=f"{api_url}/cell-morphology/{entity_id}/assets",
         match_headers=request_headers,
         json=_mock_asset_response(asset_id),
     )
@@ -555,7 +589,7 @@ def test_client_download_assets(
         | {
             "assets": [
                 _mock_asset_response(asset1_id)
-                | {"path": "foo/bar/bar.h5", "content_type": "application/hdf5"},
+                | {"path": "foo/bar/bar.h5", "content_type": "application/x-hdf5"},
                 _mock_asset_response(asset2_id)
                 | {"path": "foo/bar/bar.swc", "content_type": "application/swc"},
             ]
@@ -583,8 +617,8 @@ def test_client_download_assets(
     ).one()
 
     assert res.asset.path == "foo/bar/bar.swc"
-    assert res.output_path == tmp_path / "foo/bar/bar.swc"
-    assert res.output_path.read_bytes() == b"bar"
+    assert res.path == tmp_path / "foo/bar/bar.swc"
+    assert res.path.read_bytes() == b"bar"
 
 
 def test_client_download_assets__no_assets_raise(
@@ -670,6 +704,7 @@ def test_client_download_assets__entity(
                 full_path="/foo/asset1",
                 is_directory=False,
                 content_type="application/json",
+                label="cell_composition_summary",
                 size=1,
             ),
         ],
@@ -719,7 +754,7 @@ def test_upload_directory_by_paths(
             entity_type=Entity,
             name="test-directory",
             paths=paths,
-            label=None,
+            label="sonata_circuit",
             metadata=None,
         )
 
@@ -732,7 +767,7 @@ def test_upload_directory_by_paths(
         "full_path": "asdf",
         "id": "a370a57b-7211-4426-8046-970758ceaf68",
         "is_directory": True,
-        "label": None,
+        "label": "sonata_circuit",
         "meta": {},
         "path": "",
         "sha256_digest": None,
@@ -767,6 +802,45 @@ def test_upload_directory_by_paths(
     )
     assert res == Asset.model_validate(asset)
 
+    httpx_mock.add_response(
+        method="POST",
+        url=f"{api_url}/entity/{entity_id}/assets/directory/upload",
+        match_headers=request_headers,
+        json={
+            "asset": asset,
+            "files": {
+                "foo/bar/baz/subdir0/file1.txt": "http://upload_url0",
+            },
+        },
+    )
+
+    # have read error / exception
+    httpx_mock.add_exception(
+        httpx.ReadTimeout("Unable to read within timeout"),
+        method="PUT",
+        url="http://upload_url0",
+    )
+    httpx_mock.add_exception(
+        httpx.ReadTimeout("Unable to read within timeout"),
+        method="PUT",
+        url="http://upload_url0",
+    )
+    httpx_mock.add_exception(
+        httpx.ReadTimeout("Unable to read within timeout"),
+        method="PUT",
+        url="http://upload_url0",
+    )
+
+    with pytest.raises(EntitySDKError, match="Uploading these files failed"):
+        client.upload_directory(
+            entity_id=entity_id,
+            entity_type=Entity,
+            name="test-directory",
+            paths=paths,
+            label=None,
+            metadata=None,
+        )
+
     # have s3 upload fail:
     httpx_mock.add_response(
         method="POST",
@@ -784,7 +858,7 @@ def test_upload_directory_by_paths(
     httpx_mock.add_response(method="PUT", url="http://upload_url0", status_code=404)
     httpx_mock.add_response(method="PUT", url="http://upload_url0", status_code=404)
 
-    with pytest.raises(Exception, match="Uploading these files failed"):
+    with pytest.raises(EntitySDKError, match="Uploading these files failed"):
         client.upload_directory(
             entity_id=entity_id,
             entity_type=Entity,
@@ -829,12 +903,14 @@ def test_client_list_directory(
     assert res.files[Path("a/b/foo.txt")].name == "a/b/foo.txt"
 
 
+@pytest.mark.parametrize("max_concurrent", [1, 4])
 def test_client_download_directory_ignore_directory(
     tmp_path,
     client,
     httpx_mock,
     api_url,
     request_headers,
+    max_concurrent,
 ):
     entity_id = uuid.uuid4()
     asset_id = uuid.uuid4()
@@ -873,17 +949,20 @@ def test_client_download_directory_ignore_directory(
         asset_id=asset_id,
         output_path=tmp_path,
         ignore_directory_name=True,
+        max_concurrent=max_concurrent,
     )
     assert len(res) == 1
     assert res[0] == (tmp_path / "foo.txt").absolute()
 
 
+@pytest.mark.parametrize("max_concurrent", [1, 4])
 def test_client_download_directory(
     tmp_path,
     client,
     httpx_mock,
     api_url,
     request_headers,
+    max_concurrent,
 ):
     entity_id = uuid.uuid4()
     asset_id = uuid.uuid4()
@@ -936,4 +1015,155 @@ def test_client_download_directory(
             asset_id=asset_id,
             output_path=target,
             ignore_directory_name=False,
+            max_concurrent=max_concurrent,
         )
+
+
+@pytest.mark.parametrize("max_concurrent", [1, 4])
+def test_client_download_directory__asset(
+    tmp_path,
+    client,
+    httpx_mock,
+    api_url,
+    request_headers,
+    max_concurrent,
+):
+    entity_id = uuid.uuid4()
+    asset_id = uuid.uuid4()
+
+    date = "2025-01-01T00:00:00Z"
+
+    asset = Asset(
+        id=asset_id,
+        path="path_to_asset",
+        full_path="/circuit",
+        is_directory=True,
+        size=0,
+        content_type="application/vnd.directory",
+        label="sonata_circuit",
+    )
+
+    # for listing dirs
+    httpx_mock.add_response(
+        method="GET",
+        url=f"{api_url}/entity/{entity_id}/assets/{asset_id}/list",
+        match_headers=request_headers,
+        json={
+            "files": {
+                "foo.txt": {"name": "foo.txt", "size": 3, "last_modified": date},
+            }
+        },
+    )
+
+    # for downloading the asset
+    httpx_mock.add_response(
+        method="GET",
+        url=f"{api_url}/entity/{entity_id}/assets/{asset_id}/download?asset_path=foo.txt",
+        match_headers=request_headers,
+        text="file contents",
+    )
+
+    res = client.download_directory(
+        entity_id=entity_id,
+        entity_type=Entity,
+        asset_id=asset,
+        output_path=tmp_path,
+        ignore_directory_name=False,
+        max_concurrent=max_concurrent,
+    )
+    assert len(res) == 1
+    assert res[0] == (tmp_path / "path_to_asset/foo.txt").absolute()
+
+
+@patch("entitysdk.route.get_route_name")
+def test_client_register_asset(
+    mock_route,
+    client,
+    httpx_mock,
+    api_url,
+    request_headers,
+):
+    mock_route.return_value = "cell-morphology"
+
+    entity_id = uuid.uuid4()
+    asset_id = uuid.uuid4()
+
+    httpx_mock.add_response(
+        method="POST",
+        url=f"{api_url}/cell-morphology/{entity_id}/assets/register",
+        match_headers=request_headers,
+        json=_mock_asset_response(asset_id),
+    )
+
+    res = client.register_asset(
+        entity_id=entity_id,
+        entity_type=None,
+        name="foo.swc",
+        storage_path="path/to/foo.swc",
+        storage_type=StorageType.aws_s3_open,
+        is_directory=False,
+        content_type=ContentType.application_swc,
+        asset_label=AssetLabel.morphology,
+    )
+
+    assert res.id == asset_id
+    assert res.status == "created"
+
+
+@patch("entitysdk.route.get_route_name")
+def test_client_get_entity_derivations(mock_route, client, httpx_mock, api_url, request_headers):
+    mock_route.return_value = "circuit"
+    entity_id = uuid.uuid4()
+
+    derivation_1 = uuid.uuid4()
+    derivation_2 = uuid.uuid4()
+
+    used_id = uuid.uuid4()
+    generated_id = uuid.uuid4()
+
+    def add_response(derivation_type: DerivationType):
+        httpx_mock.add_response(
+            method="GET",
+            url=f"{api_url}/circuit/{entity_id}/derived-from?derivation_type={derivation_type}",
+            match_headers=request_headers,
+            json={
+                "data": [
+                    {
+                        "id": str(derivation_1),
+                        "used_id": str(used_id),
+                        "generated_id": str(generated_id),
+                        "derivation_type": derivation_type,
+                    },
+                    {
+                        "id": str(derivation_2),
+                        "used_id": str(used_id),
+                        "generated_id": str(generated_id),
+                        "derivation_type": derivation_type,
+                    },
+                ]
+            },
+        )
+
+    add_response(DerivationType.circuit_extraction)
+    res = client.get_entity_derivations(
+        entity_id=entity_id, entity_type=Circuit, derivation_type=DerivationType.circuit_extraction
+    ).all()
+    assert len(res) == 2
+    assert res[0].id == derivation_1
+    assert res[1].id == derivation_2
+
+    add_response(DerivationType.circuit_rewiring)
+    res = client.get_entity_derivations(
+        entity_id=entity_id, entity_type=Circuit, derivation_type=DerivationType.circuit_rewiring
+    ).all()
+    assert len(res) == 2
+    assert res[0].id == derivation_1
+    assert res[1].id == derivation_2
+
+    add_response(DerivationType.unspecified)
+    res = client.get_entity_derivations(
+        entity_id=entity_id, entity_type=Circuit, derivation_type=DerivationType.unspecified
+    ).all()
+    assert len(res) == 2
+    assert res[0].id == derivation_1
+    assert res[1].id == derivation_2
